// kmidentity.cpp

#include "kmidentity.h"
#include "kfileio.h"

#include <kapp.h>
#include <kconfig.h>

#include <pwd.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include "extern.h"
#include "kpasswdconfig.h"

#ifndef _PATH_SENDMAIL
#define _PATH_SENDMAIL  "/usr/sbin/sendmail"
#endif

//-----------------------------------------------------------------------------
KMIdentity::KMIdentity(const char *virt)
{
	KConfig *config = kapp->getConfig();
	mCurrentGroup = config->groupIterator();
	
	setIdentity(virt);
  readConfig();
}


//-----------------------------------------------------------------------------
KMIdentity::~KMIdentity()
{
	delete mCurrentGroup;
	mCurrentGroup = NULL;
  writeConfig();
}

//-----------------------------------------------------------------------------
void KMIdentity::setIdentity(const char *virt)
{
	mIdentity = "Identity";
	if(virt) mIdentity = mIdentity + "-" + virt;
	mCurrentGroup->toFirst();
	while(mCurrentGroup->current())
	{
		if(mIdentity == mCurrentGroup->currentKey()) break;
		mCurrentGroup->operator++();
	}
  readConfig();
}

//-----------------------------------------------------------------------------
void KMIdentity::setIdentityFor(const char *folder)
{
	setIdentity(NULL);
}


//-----------------------------------------------------------------------------
void KMIdentity::readConfig(void)
{
  KConfig* config = kapp->getConfig();
  struct passwd* pw;
  char str[80];
  int i;

  config->setGroup(mIdentity);
  mFullName = config->readEntry("Name");
  if (mFullName.isEmpty())
  {
    pw = getpwuid(getuid());
    if (pw)
    {
      mFullName = pw->pw_gecos;
      mFullName.detach();
      i = mFullName.find(',');
      if (i>0) mFullName.truncate(i);
    }
  }


  mEmailAddr = config->readEntry("Email Address");
  if (mEmailAddr.isEmpty())
  {
    pw = getpwuid(getuid());
    if (pw)
    {
      gethostname(str, 79);
      mEmailAddr = QString(pw->pw_name) + "@" + str;
      mEmailAddr.detach();
    }
  }

  mOrganization = config->readEntry("Organization");
  mReplyToAddr = config->readEntry("Reply-To Address");
  mDefaultKey = config->readEntry("PGP Default Key");
  mSignatureFile = config->readEntry("Signature File");

  mSendImmediate = (bool)config->readNumEntry("Immediate", TRUE);
  mSendQuotedPrintable = (bool)config->readNumEntry("Quoted-Printable", FALSE);
  mMailer = config->readEntry("Mailer",  _PATH_SENDMAIL);
  mSmtpHost = config->readEntry("Smtp Host", "localhost");
  mSmtpPort = config->readNumEntry("Smtp Port", 25);
  mSaveSend = config->readEntry("Save Send");

  QString temp = config->readEntry("Method");
  if (temp=="mail") mMethod = smMail;
  else if (temp=="smtp") mMethod = smSMTP;
  else mMethod = smUnknown;
}


//-----------------------------------------------------------------------------
void KMIdentity::writeConfig(bool aWithSync)
{
  KConfig* config = kapp->getConfig();
  config->setGroup(mIdentity);

  config->writeEntry("Name", mFullName);
  config->writeEntry("Organization", mOrganization);
  config->writeEntry("Email Address", mEmailAddr);
  config->writeEntry("Reply-To Address", mReplyToAddr);
  config->writeEntry("PGP Default Key", mDefaultKey);
  config->writeEntry("Signature File", mSignatureFile);

  config->writeEntry("Immediate", mSendImmediate);
  config->writeEntry("Quoted-Printable", mSendQuotedPrintable);
  config->writeEntry("Mailer", mMailer);
  config->writeEntry("Smtp Host", mSmtpHost);
  config->writeEntry("Smtp Port", mSmtpPort);
  config->writeEntry("Save Send", mSaveSend);
  config->writeEntry("Method", (mMethod==smSMTP) ? "smtp" : "mail");

  if (aWithSync) config->sync();
}


//-----------------------------------------------------------------------------
bool KMIdentity::mailingAllowed(void) const
{
  if (mMethod!=smSMTP && mMethod!=smMail)
  {
    warning(i18n("Please specify a send\nmethod in the settings\n"
			   "and try again."));
    return FALSE;
  }
  return (!mFullName.isEmpty() && !mEmailAddr.isEmpty());
}


//-----------------------------------------------------------------------------
void KMIdentity::setFullName(const QString str)
{
  mFullName = str.copy();
}


//-----------------------------------------------------------------------------
void KMIdentity::setOrganization(const QString str)
{
  mOrganization = str.copy();
}

//-----------------------------------------------------------------------------
void KMIdentity::setPgpDefaultKey(const QString str)
{
  mDefaultKey = str.copy();
	passwd->setKey(str);
}


//-----------------------------------------------------------------------------
void KMIdentity::setEmailAddr(const QString str)
{
  mEmailAddr = str.copy();
}


//-----------------------------------------------------------------------------
const QString KMIdentity::fullEmailAddr(void) const
{
  QString result;

  if (mFullName.isEmpty()) result = mEmailAddr.copy();
  else result = mFullName.copy() + " <" + mEmailAddr + ">";

  return result;
}

//-----------------------------------------------------------------------------
void KMIdentity::setReplyToAddr(const QString str)
{
  mReplyToAddr = str.copy();
}


//-----------------------------------------------------------------------------
void KMIdentity::setSignatureFile(const QString str)
{
  mSignatureFile = str.copy();
}


//-----------------------------------------------------------------------------
const QString KMIdentity::signature(void) const
{
  QString result, sigcmd;
  char tmpf[256];

  if (mSignatureFile.isEmpty()) return 0;

  if (mSignatureFile.right(1)=="|")
  {
    // signature file is a shell script that returns the signature
    tmpnam(tmpf);
    sigcmd = mSignatureFile.left(mSignatureFile.length()-1);
    sigcmd += " >";
    sigcmd += tmpf;
    system(sigcmd);

    if (errno)
    {
      warning(i18n("Failed to execute signature script\n%s\n%s"),
	      sigcmd.data(), strerror(errno));
      return 0;
    }
    result = kFileToString(tmpf, TRUE, FALSE);
    unlink(tmpf);
  }
  else
  {
    result = kFileToString(mSignatureFile);
  }

  return result;
}

//-----------------------------------------------------------------------------
void KMIdentity::setMethod(Method aMethod)
{
  mMethod = aMethod;
}


//-----------------------------------------------------------------------------
void KMIdentity::setSendImmediate(bool aSendImmediate)
{
  mSendImmediate = aSendImmediate;
}


//-----------------------------------------------------------------------------
void KMIdentity::setSendQuotedPrintable(bool aSendQuotedPrintable)
{
  mSendQuotedPrintable = aSendQuotedPrintable;
}


//-----------------------------------------------------------------------------
void KMIdentity::setMailer(const QString& aMailer)
{
  mMailer = aMailer;
}


//-----------------------------------------------------------------------------
void KMIdentity::setSmtpHost(const QString& aSmtpHost)
{
  mSmtpHost = aSmtpHost;
}


//-----------------------------------------------------------------------------
void KMIdentity::setSmtpPort(unsigned short int aSmtpPort)
{
  mSmtpPort = aSmtpPort;
}

//-----------------------------------------------------------------------------
void KMIdentity::setSaveSend(const QString& aSaveSend)
{
  mSaveSend = aSaveSend;
}

bool KMIdentity::setNext()
{
	while(mCurrentGroup->current())
	{
		mCurrentGroup->operator++();
		if(mCurrentGroup->current() && !strncmp("Identity",mCurrentGroup->currentKey(),8))
		{
 			if(strlen(mCurrentGroup->currentKey()) > 9) setIdentity(&mCurrentGroup->currentKey()[9]);
			break;
		}		
	}
	return(mCurrentGroup->current() != NULL);
}

bool KMIdentity::toFirst()
{
	mCurrentGroup->toFirst();
	while(mCurrentGroup->current())
	{
		if(mCurrentGroup->current() && !strncmp("Identity",mCurrentGroup->currentKey(),8))
		{
 			if(strlen(mCurrentGroup->currentKey()) > 9) setIdentity(&mCurrentGroup->currentKey()[9]);
			break;
		}		
		mCurrentGroup->operator++();
	}
	return(mCurrentGroup->current() != NULL);
}
/*	
	KConfig *config=kapp->getConfig();
	KGroupIterator *git=config->groupIterator();
	
	if(git) {
*/