#include "kpgpbase.h"
#include "kpgp.h"

#include <kapp.h>
#include <ksimpleconfig.h>

#include <qregexp.h>

#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <stdlib.h>
#include <errno.h>
#include <klocale.h>
#include <extern.h>
#include <sys/time.h>
#include "kimap.h"
//#include "nmstatusbar.h"

QString QStringnull()
{
	QString empty;
	return empty;
}

KpgpBase::KpgpBase()
: QObject()
{
  readConfig();
  status = pgpOK;
}

KpgpBase::~KpgpBase()
{
  
}

void 
KpgpBase::readConfig()
{
  KSimpleConfig *config = Kpgp::getConfig();
  pgpUser = config->readEntry("user");
  flagEncryptToSelf = config->readBoolEntry("encryptToSelf");
}

void
KpgpBase::writeConfig(bool sync)
{
  KSimpleConfig *config = Kpgp::getConfig();
  config->writeEntry("user",pgpUser);
  config->writeEntry("encryptToSelf",flagEncryptToSelf);

  if(sync)
    config->sync();
}


bool
KpgpBase::setMessage(const QString mess)
{
  clear();
  input = mess;
  if(input.find("-----BEGIN PGP") != -1)
    decrypt();
}

QString 
KpgpBase::message() const
{
  // do we have a deciphered text?
  if(!output.isEmpty()) return output;

  // no, then return the original one
  //debug("KpgpBase: No output!");
  return input;
}

int
KpgpBase::run(const char *cmd, const char *passphrase)
{
  /* the pipe ppass is used for to pass the password to
   * pgp. passing the password together with the normal input through
   * stdin doesn't seem to work as expected (at least for pgp5.0)
   */
  int ppass[2];
  FILE *pass;

  if(passphrase)
  {
    pipe(ppass);

    pass = fdopen(ppass[1], "w");
    fwrite(passphrase, sizeof(char), strlen(passphrase), pass);
    fwrite("\n", sizeof(char), 1, pass);
    fclose(pass);
    close(ppass[1]);

    // tell pgp which fd to use for the passphrase
    QString tmp;
    tmp.sprintf("%d",ppass[0]);
    setenv("PGPPASSFD",tmp,1);

    //printf("PGPPASSFD = %s\n",tmp.data());
    //printf("pass = %s\n",passphrase);
  }
  else
    unsetenv("PGPPASSFD");

  //printf("cmd = %s\n",cmd);
  //printf("input = %s\nlength = %d\n",input.data(), input.length());

	runGeneric(cmd);
  unsetenv("PGPPASSFD");
  return pgpOK;
}

int
KpgpBase::runGpg(const char *cmd, const char *passphrase)
{
  /* the pipe ppass is used for to pass the password to
   * pgp. passing the password together with the normal input through
   * stdin doesn't seem to work as expected (at least for pgp5.0)
   */
  int ppass[2];
  FILE *pass;
  char gpgcmd[1024] = "\0";

  if(passphrase)
  {
    pipe(ppass);

    pass = fdopen(ppass[1], "w");
    fwrite(passphrase, sizeof(char), strlen(passphrase), pass);
    fwrite("\n", sizeof(char), 1, pass);
    fclose(pass);
    close(ppass[1]);

    //printf("pass = %s\n",passphrase);
  }

  //printf("cmd = %s\n",cmd);
  //printf("input = %s\nlength = %d\n",input.data(), input.length());


  if(passphrase) {
    snprintf(gpgcmd, 1023, "gpg --passphrase-fd %d %s",ppass[0],cmd);
  } else {
    snprintf(gpgcmd, 1023, "gpg %s",cmd);
  }
	
	runGeneric(gpgcmd);
	
	//close the rest of the streams
  if(passphrase)
    close(ppass[0]);
	
  return pgpOK;
}


QString
KpgpBase::addUserId()
{
  QString cmd;

  if(!pgpUser.isEmpty())
  {
    cmd += " -u \"";
    cmd += pgpUser;
    cmd += "\"";
    return cmd;
  }
  return "";
}

void 
KpgpBase::clear()
{
  input = QStringnull();
  output = QStringnull();
  info = QStringnull();
  errMsg = QStringnull();
  signature = QStringnull();
  signatureID = QStringnull();
  recipients.clear();
  status = pgpOK;
}

void 
KpgpBase::clearOutput()
{
  output = QStringnull();
}

QString 
KpgpBase::lastErrorMessage() const
{
  return errMsg;
}

// -------------------------------------------------------------------------

KpgpBaseG::KpgpBaseG()
  : KpgpBase()
{
}

KpgpBaseG::~KpgpBaseG()
{
}

int 
KpgpBaseG::encrypt(const QStrList *_recipients, bool /*ignoreUntrusted*/)
{
  return encsign(_recipients, 0);
}

int 
KpgpBaseG::sign(const char *passphrase)
{
  return encsign(0, passphrase);
}

int 
KpgpBaseG::encsign(const QStrList *_recipients, const char *passphrase,
		   bool /*ignoreUntrusted*/)
{
  QString cmd, pers;
  output = "";

  if(_recipients != 0)
    if(_recipients->count() <= 0)
      _recipients = 0;

  if(_recipients != 0 && passphrase != 0)
    cmd = "--batch --escape-from --armor --always-trust --sign --encrypt ";
  else if( _recipients != 0 )
    cmd = "--batch --escape-from --armor --always-trust --encrypt ";
  else if(passphrase != 0 )
    cmd = "--batch --escape-from --armor --always-trust --clearsign ";
  else if(flagEncryptToSelf)
    cmd = "--batch --escape-from --armor --always-trust --encrypt ";
  else
  {
    debug("kpgpbase: Neither recipients nor passphrase specified.");
    return pgpOK;
  }

  if(passphrase != 0)
    cmd += addUserId();

  if(_recipients != 0)
  {
    QStrListIterator it(*_recipients);
    while( (pers=it.current()) )
    {
      cmd += " --recipient \"";
      cmd += pers;
      cmd += "\" ";
      ++it;
    }
  }
  if(flagEncryptToSelf)
    cmd += " --default-recipient-self";
  cmd += " --set-filename stdin ";

  status = runGpg(cmd, passphrase);
  if(status == pgpRUN_ERR) return status;

  if(_recipients != 0 || flagEncryptToSelf)
  {
    int index = 0;
    bool bad = FALSE;
    unsigned int num = 0;
    QString badkeys = "";
    while((index = info.find("Cannot find the public key",index)) 
	  != -1)
    {
      bad = TRUE;
      index = info.find("'",index);
      int index2 = info.find("'",index+1);
      badkeys += info.mid(index, index2-index+1) + ", ";
      num++;
    }
    if(bad)
    {
      badkeys.stripWhiteSpace();
      if(num == _recipients->count())
	errMsg.sprintf("Could not find public keys matching the\n" 
		       "userid(s) %s.\n" 
		       "Message is not encrypted.\n",
		       (const char *)badkeys);
      else
	errMsg.sprintf("Could not find public keys matching the\n" 
		       "userid(s) %s. These persons won't be able\n"
		       "to read the message.",
		       (const char *)badkeys);
      status |= pgpMISSINGKEY;
      status |= pgpERROR;
    }
  }
  if(passphrase != 0)
  {
    if(info.find("Pass phrase is good") != -1)
    {
      //debug("KpgpBase: Good Passphrase!");
      status |= pgpSIGNED;
    }
    if( info.find("Bad pass phrase") != -1)
    {
      errMsg = i18n("Bad pass Phrase; couldn't sign");
      status |= pgpERR_SIGNING;
      status |= pgpERROR;
    }
  }
  //debug("status = %d",status);
  return status;
}

int
KpgpBaseG::verifySig(const char *signature_file)
{
  QString cmd;
  int index, index2;
  output = "";

  cmd.sprintf("--batch --set-filename stdin --verify %s",signature_file);
  status = runGpg(cmd, NULL);

  if((index = info.find("Signature made")) != -1)
  {
    //debug("KpgpBase: message is signed");
    status |= pgpSIGNED;
    if( info.find("Key matching expected") != -1)
    {
      index = info.find("Key ID ",index);
      signatureID = info.mid(index+7,8);
      signature = i18n("unknown key ID ") + signatureID + " ";
      status |= pgpUNKNOWN_SIG;
      status |= pgpGOODSIG;
    }
    else if( info.find("Good signature") != -1 )
    {
      status |= pgpGOODSIG;
      // get signer
      index = info.find("\"",index);
      index2 = info.find("\"", index+1);
      signature = info.mid(index+1, index2-index-1);

      // get key ID of signer
      index = info.find("key ID ",index2);
      signatureID = info.mid(index+7,8);
    }
    else if( info.find("BAD signature") != -1 )
    {
      // get signer
      index = info.find("\"",index);
      index2 = info.find("\"", index+1);
      signature = info.mid(index+1, index2-index-1);

      // get key ID of signer
      index = info.find("key ID ",index2);
      signatureID = info.mid(index+7,8);
    }
    else if( info.find("CRC error") != -1 )
    {
      //debug("BAD signature");
      status |= pgpSIGNED;
      status |= pgpERROR;

      // get signer
      index = info.find("\"",index);
      index2 = info.find("\"", index+1);
      signature = info.mid(index+1, index2-index-1);

      // get key ID of signer
      index = info.find("key ID ",index2);
      signatureID = info.mid(index+7,8);
    }
    else if( info.find("Can't find the right public key") != -1 )
    {
      status |= pgpUNKNOWN_SIG;
      status |= pgpGOODSIG; // this is a hack...
      signature = i18n("??? (file ~/.gnupg/pubring.gpg not found)");
      signatureID = "???";
    }
    else if( info.find("Can't check signature") != -1 )
    {
			int i,j;
      status |= pgpUNKNOWN_SIG;
      status |= pgpGOODSIG; // this is a hack...

			i = info.find("key ID ") + 7;
			j = info.find(0x0d,i) ? info.find(0x0a,i) : info.find(0x0d,i);
			if(i<j) signatureID = info.mid(i,j-i);
      else signatureID = "???";

			i = info.find("Can't check signature") + 23;
			j = info.find(0x0d,i) ? info.find(0x0a,i) : info.find(0x0d,i);
			if(i<j) signature = signatureID + " (" + i18n(info.mid(i,j-i)) + ")";
      else signature = signatureID + " (???)";
    }
    else
    {
      status |= pgpERROR;
      signature = "";
      signatureID = "";
			fprintf(stderr,"PGPINFO: '%s'\n",info.data());
    }
  }
  else
  {
    status |= pgpERROR;
    signature = "";
    signatureID = "";
		fprintf(stderr,"PGPINFO: '%s'\n",info.data());
  }
  debug("status = %d",status);

	return status;
}

int 
KpgpBaseG::decrypt(const char *passphrase)
{
  QString cmd;
  int index, index2;
  output = "";

  cmd = "--batch --set-filename stdin --decrypt";

  status = runGpg(cmd, passphrase);

  // pgp2.6 has sometimes problems with the ascii armor pgp5.0 produces
  // this hack can solve parts of the problem
  if(info.find("ASCII armor corrupted.") != -1)
  {
    debug("removing ASCII armor header");
    int index1 = input.find("-----BEGIN PGP SIGNED MESSAGE-----");
    if(index1 != -1)
      index1 = input.find("-----BEGIN PGP SIGNATURE-----", index1);
    else
      index1 = input.find("-----BEGIN PGP MESSAGE-----");
    index1 = input.find("\n", index1);
    index2 = input.find("\n\n", index1);
    input.remove(index1, index2 - index1);
    status = runGpg(cmd, passphrase);
  }
 
  if(status == pgpRUN_ERR)
  {
    errMsg = i18n("error running gpg");
    return status;
  }

  if( info.find("File contains key") != -1)
  {
    // FIXME: should do something with it...
  }

  if(info.find("secret key not available") != -1)
  {
    //debug("kpgpbase: message is encrypted");
    status |= pgpENCRYPTED;
    if( info.find("bad passphrase") != -1)
    {
      if(passphrase != 0) 
      {
				errMsg = i18n("Bad pass Phrase; couldn't decrypt");
				debug("KpgpBase: passphrase is bad");
				status |= pgpBADPHRASE;
				status |= pgpERROR;
      }
    } 
    else
    {
      // no secret key fitting this message
      status |= pgpNO_SEC_KEY;
      status |= pgpERROR;
      errMsg = i18n("Do not have the secret key for this message");
      debug("KpgpBase: no secret key for this message");
    }
    // check for persons
    index = info.find("can only be read by:");
    if(index != -1) 
    {
      index = info.find("\n",index);
      int end = info.find("\n\n",index);
      
      recipients.clear();
      while( (index2 = info.find("\n",index+1)) <= end )
      {
				QString item = info.mid(index+1,index2-index-1);
				item.stripWhiteSpace();
				recipients.append(item);
				index = index2;
      }
    }
  }
  if((index = info.find("Signature made")) != -1)
  {
    //debug("KpgpBase: message is signed");
    status |= pgpSIGNED;
    if( info.find("Key matching expected") != -1)
    {
      index = info.find("Key ID ",index);
      signatureID = info.mid(index+7,8);
      signature = i18n("unknown key ID ") + signatureID + " ";
      status |= pgpUNKNOWN_SIG;
      status |= pgpGOODSIG;
    }
    else if( info.find("Good signature") != -1 )
    {
      status |= pgpGOODSIG;
      // get signer
      index = info.find("\"",index);
      index2 = info.find("\"", index+1);
      signature = info.mid(index+1, index2-index-1);
      
      // get key ID of signer
      index = info.find("key ID ",index2);
      signatureID = info.mid(index+7,8);
    }
    else if( info.find("BAD signature") != -1 )
    {
      // get signer
      index = info.find("\"",index);
      index2 = info.find("\"", index+1);
      signature = info.mid(index+1, index2-index-1);

      // get key ID of signer
      index = info.find("key ID ",index2);
      signatureID = info.mid(index+7,8);
    }
    else if( info.find("CRC error") != -1 )
    {
      //debug("BAD signature");
      status |= pgpSIGNED;
      status |= pgpERROR;

      // get signer
      index = info.find("\"",index);
      index2 = info.find("\"", index+1);
      signature = info.mid(index+1, index2-index-1);
      
      // get key ID of signer
      index = info.find("key ID ",index2);
      signatureID = info.mid(index+7,8);
    }
    else if( info.find("Can't find the right public key") != -1 )
    {
      status |= pgpUNKNOWN_SIG;
      status |= pgpGOODSIG; // this is a hack...
      signature = i18n("??? (file ~/.gnupg/pubring.gpg not found)");
      signatureID = "???";
    }
    else if( info.find("Can't check signature") != -1 )
    {
			int i,j;
      status |= pgpUNKNOWN_SIG;
      status |= pgpGOODSIG; // this is a hack...

			i = info.find("key ID ") + 7;
			j = info.find(0x0d,i) ? info.find(0x0a,i) : info.find(0x0d,i);
			if(i<j) signatureID = info.mid(i,j-i);
      else signatureID = "???";

			i = info.find("Can't check signature") + 23;
			j = info.find(0x0d,i) ? info.find(0x0a,i) : info.find(0x0d,i);
			if(i<j) signature = signatureID + " (" + i18n(info.mid(i,j-i)) + ")";
      else signature = signatureID + " (???)";
    }
    else
    {
      status |= pgpERROR;
      signature = "";
      signatureID = "";
			fprintf(stderr,"PGPINFO: '%s'\n",info.data());
    }
  }
  debug("status = %d",status);
  debug("PGP info = %s",info.data());
	return status;
}

QStrList
KpgpBaseG::pubKeys()
{
  QString cmd;
  int index, index2;

  cmd = "--batch --list-keys";
  status = runGpg(cmd);
  if(status == pgpRUN_ERR) return 0;

  // now we need to parse the output
  QStrList publicKeys;
  index = output.find("\n",1)+1; // skip first to "\n"
  while( (index = output.find("\n",index)) != -1)
  {
    //parse line
    QString line;
    if( (index2 = output.find("\n",index+1)) != -1)
    {
      int index3 = output.find("pub ",index);
		    
      if( (index3 <index2) && (index3 != -1) )
      {
	line = output.mid(index3+31,index2-index3-31);
	line = line.lower();
      }
      if(!line.isEmpty())
      {
	//debug("KpgpBase: found key for %s.",(const char *)line);
	publicKeys.append(line);
      }
    }
    index = index2;
  }
  //debug("finished reading keys");
  return publicKeys;
}

int
KpgpBaseG::signKey(const char *key, const char *passphrase)
{
  QString cmd;

  cmd = "--set-filename stdin ";
  cmd += addUserId();
  cmd += "--sign-key ";
  cmd += key;

  return runGpg(cmd,passphrase);
}


QString KpgpBaseG::getAsciiPublicKey(QString _person) {

  QString toexec;
  toexec.sprintf("--batch --armor --export \"%s\"", _person.data());

  status = runGpg(toexec.data());
  if(status == pgpRUN_ERR) return 0;

  return output;
}


// -------------------------------------------------------------------------

KpgpBase2::KpgpBase2()
  : KpgpBase()
{
}

KpgpBase2::~KpgpBase2()
{
}

int 
KpgpBase2::encrypt(const QStrList *_recipients, bool /*ignoreUntrusted*/)
{
  return encsign(_recipients, 0);
}

int 
KpgpBase2::sign(const char *passphrase)
{
  return encsign(0, passphrase);
}

int 
KpgpBase2::encsign(const QStrList *_recipients, const char *passphrase,
		   bool /*ignoreUntrusted*/)
{
  QString cmd, pers;
  output = "";

  if(_recipients != 0)
    if(_recipients->count() <= 0)
      _recipients = 0;

  if(_recipients != 0 && passphrase != 0)
    cmd = "pgp +batchmode -seat ";
  else if( _recipients != 0 )
    cmd = "pgp +batchmode -eat";
  else if(passphrase != 0 )
    cmd = "pgp +batchmode -sat ";
  else if(flagEncryptToSelf)
    cmd = "pgp +batchmode -eat ";
  else
  {
    debug("kpgpbase: Neither recipients nor passphrase specified.");
    return pgpOK;
  }

  if(passphrase != 0)
    cmd += addUserId();

  if(_recipients != 0)
  {
    QStrListIterator it(*_recipients);
    while( (pers=it.current()) )
    {
      cmd += " \"";
      cmd += pers;
      cmd += "\"";
      ++it;
    }
  }
  if(flagEncryptToSelf)
    cmd += " +EncryptToSelf";
  cmd += " -f";

  status = run(cmd, passphrase);
  if(status == pgpRUN_ERR) return status;

  if(_recipients != 0)
  {
    int index = 0;
    bool bad = FALSE;
    unsigned int num = 0;
    QString badkeys = "";
    while((index = info.find("Cannot find the public key",index)) 
	  != -1)
    {
      bad = TRUE;
      index = info.find("'",index);
      int index2 = info.find("'",index+1);
      badkeys += info.mid(index, index2-index+1) + ", ";
      num++;
    }
    if(bad)
    {
      badkeys.stripWhiteSpace();
      if(num == _recipients->count())
	errMsg.sprintf("Could not find public keys matching the\n" 
		       "userid(s) %s.\n" 
		       "Message is not encrypted.\n",
		       (const char *)badkeys);
      else
	errMsg.sprintf("Could not find public keys matching the\n" 
		       "userid(s) %s. These persons won't be able\n"
		       "to read the message.",
		       (const char *)badkeys);
      status |= pgpMISSINGKEY;
      status |= pgpERROR;
    }
  }
  if(passphrase != 0)
  {
    if(info.find("Pass phrase is good") != -1)
    {
      //debug("KpgpBase: Good Passphrase!");
      status |= pgpSIGNED;
    }
    if( info.find("Bad pass phrase") != -1)
    {
      errMsg = i18n("Bad pass Phrase; couldn't sign");
      status |= pgpERR_SIGNING;
      status |= pgpERROR;
    }
  }
  //debug("status = %d",status);
  return status;
}

int 
KpgpBase2::decrypt(const char *passphrase)
{
  QString cmd;
  int index, index2;
  output = "";


  cmd = "pgp +batchmode -f";

  status = run(cmd, passphrase);

  // pgp2.6 has sometimes problems with the ascii armor pgp5.0 produces
  // this hack can solve parts of the problem
  if(info.find("ASCII armor corrupted.") != -1)
  {
    debug("removing ASCII armor header");
    int index1 = input.find("-----BEGIN PGP SIGNED MESSAGE-----");
    if(index1 != -1)
      index1 = input.find("-----BEGIN PGP SIGNATURE-----", index1);
    else
      index1 = input.find("-----BEGIN PGP MESSAGE-----");
    index1 = input.find("\n", index1);
    index2 = input.find("\n\n", index1);
    input.remove(index1, index2 - index1);
    status = run(cmd, passphrase);
  }
 
  if(status == pgpRUN_ERR)
  {
    errMsg = i18n("error running pgp");
    return status;
  }

  if( info.find("File contains key") != -1)
  {
    // FIXME: should do something with it...
  }

  if(info.find("You do not have the secret key") != -1)
  {
    //debug("kpgpbase: message is encrypted");
    status |= pgpENCRYPTED;
    if( info.find("Bad pass phrase") != -1)
    {
      if(passphrase != 0) 
      {
	errMsg = i18n("Bad pass Phrase; couldn't decrypt");
	debug("KpgpBase: passphrase is bad");
	status |= pgpBADPHRASE;
	status |= pgpERROR;
      }
    } 
    else
    {
      // no secret key fitting this message
      status |= pgpNO_SEC_KEY;
      status |= pgpERROR;
      errMsg = i18n("Do not have the secret key for this message");
      debug("KpgpBase: no secret key for this message");
    }
    // check for persons
    index = info.find("can only be read by:");
    if(index != -1) 
    {
      index = info.find("\n",index);
      int end = info.find("\n\n",index);
      
      recipients.clear();
      while( (index2 = info.find("\n",index+1)) <= end )
      {
	QString item = info.mid(index+1,index2-index-1);
	item.stripWhiteSpace();
	recipients.append(item);
	index = index2;
      }
    }
  }
  if((index = info.find("File has signature")) != -1)
  {
    //debug("KpgpBase: message is signed");
    status |= pgpSIGNED;
    if( info.find("Key matching expected") != -1)
    {
      index = info.find("Key ID ",index);
      signatureID = info.mid(index+7,8);
      signature = i18n("unknown key ID ") + signatureID + " ";
      status |= pgpUNKNOWN_SIG;
      status |= pgpGOODSIG;
    }
    else if( info.find("Good signature") != -1 )
    {
      status |= pgpGOODSIG;
      // get signer
      index = info.find("\"",index);
      index2 = info.find("\"", index+1);
      signature = info.mid(index+1, index2-index-1);
      
      // get key ID of signer
      index = info.find("key ID ",index2);
      signatureID = info.mid(index+7,8);
    }
    else if( info.find("Can't find the right public key") != -1 )
    {
      status |= pgpUNKNOWN_SIG;
      status |= pgpGOODSIG; // this is a hack...
      signature = i18n("??? (file ~/.pgp/pubring.pgp not found)");
      signatureID = "???";
    }
    else
    {
      status |= pgpERROR;
      signature = "";
      signatureID = "";
    }
  }
  //debug("status = %d",status);
  return status;
}

QStrList
KpgpBase2::pubKeys()
{
  QString cmd;
  int index, index2;

  cmd = "pgp +batchmode -kv -f";
  status = run(cmd);
  if(status == pgpRUN_ERR) return 0;

  QStrList publicKeys;
  index = output.find("\n",1)+1; // skip first to "\n"
  while( (index = output.find("\n",index)) != -1)
  {
    //parse line
    QString line;
    if( (index2 = output.find("\n",index+1)) != -1)
      // skip last line
    {
      int index3 = output.find("pub ",index);
      
      if( (index3 >index2) || (index3 == -1) )
      {
	// second adress for the same key
	line = output.mid(index+1,index2-index-1);
	line = line.stripWhiteSpace();	       
	line = line.lower();
      } else {
	// line with new key
	int index3 = output.find(
	  QRegExp("/[0-9][0-9]/[0-9][0-9] "),
	  index);
	line = output.mid(index3+7,index2-index3-7);
	line = line.lower();
      }
      //debug("KpgpBase: found key for %s",(const char *)line);
      publicKeys.append(line);
    }
    index = index2;
  }
  
  return publicKeys;
}

int
KpgpBase2::signKey(const char *key, const char *passphrase)
{
  QString cmd;

  cmd = "pgp +batchmode -ks -f";
  cmd += addUserId();
  if(passphrase != 0)
  {
    QString aStr;
    aStr.sprintf(" \"-z%s\"",passphrase);
    cmd += aStr;
  }
  cmd += key;

  return run(cmd);
}


QString KpgpBase2::getAsciiPublicKey(QString _person) {

  QString toexec;
  toexec.sprintf("pgp -kxaf \"%s\"", _person.data());

  status = run(toexec.data());
  if(status == pgpRUN_ERR) return QStringnull();

  return output;
}


// -------------------------------------------------------------------------

KpgpBase5::KpgpBase5()
  : KpgpBase()
{
}
KpgpBase5::~KpgpBase5()
{
}

int 
KpgpBase5::encrypt(const QStrList *_recipients, bool ignoreUntrusted)
{
  return encsign(_recipients, 0, ignoreUntrusted);
}

int 
KpgpBase5::sign(const char *passphrase)
{
  return encsign( 0, passphrase);
}
  
int 
KpgpBase5::encsign(const QStrList *_recipients, const char *passphrase,
		   bool ignoreUntrusted)
{
  QString in,cmd,pers;
  int index;
  // used to work around a bug in pgp5. pgp5 treats files
  // with non ascii chars (umlauts, etc...) as binary files, but
  // we wan't a clear signature
  bool signonly = false;
  
  output = "";
  
  if(_recipients != 0)
    if(_recipients->isEmpty())
      _recipients = 0;

  if(_recipients != 0 && passphrase != 0)
    cmd = "pgpe -ats -f +batchmode=1";
  else if( _recipients != 0 )
    cmd = "pgpe -at -f +batchmode=1 ";
  else if(passphrase != 0 )
  {
    cmd = "pgps -bat -f +batchmode=1 ";
    signonly = true;
  }
  else if(flagEncryptToSelf)
    cmd = "pgpe -at -f +batchmode=1 ";
  else
  {
    errMsg = "Neither recipients nor passphrase specified.";
    return pgpOK;
  }

  if(ignoreUntrusted) cmd += " +NoBatchInvalidKeys=off";

  if(passphrase != 0)
    cmd += addUserId();

  if(_recipients != 0)
  {
    QStrListIterator it(*_recipients);
    while( (pers=it.current()) )
    {
      cmd += " -r \"";
      cmd += pers;
      cmd += "\"";
      ++it;
    }
  }
  if(flagEncryptToSelf)
    cmd += " +EncryptToSelf";

  status = run(cmd, passphrase);
  if(status == pgpRUN_ERR) return status;
  
  // now parse the returned info
  if(info.find("Cannot unlock private key") != -1)
  {
    errMsg = i18n("The passphrase you entered is invalid.");
    status |= pgpERROR;
    status |= pgpBADPHRASE;
  }
  if(!ignoreUntrusted)
  {
    QString aStr;
    index = -1;
    while((index = info.find("WARNING: The above key",index+1)) != -1)
    {
      int index2 = info.find("But you previously",index);
      int index3 = info.find("WARNING: The above key",index+1);
      if(index2 == -1 || (index2 > index3 && index3 != -1))
      {
	// the key wasn't valid, no encryption to this person
	// extract the person
	index2 = info.find("\n",index);
	index3 = info.find("\n",index2+1);
	aStr += info.mid(index2+1, index3-index2-1);
	aStr += ", ";
      }
    }
    if(!aStr.isEmpty())
    {
      aStr.truncate(aStr.length()-2);
      if(info.find("No valid keys found") != -1)
	errMsg = i18n("The key(s) you wan't to encrypt your message\n"
		      "to are not trusted. No encryption done.");
      else
	errMsg = i18n("The following key(s) are not trusted:\n%1\n"
			    "They will not be able to decrypt the message");
//		       .arg(aStr);
      status |= pgpERROR;
      status |= pgpBADKEYS;
    }
  }
  if((index = info.find("No encryption keys found for")) != -1)
  {
    index = info.find(":",index);
    int index2 = info.find("\n",index);

    errMsg.sprintf("Missing encryption key(s) for: %s", 
		   info.mid(index,index2-index).data());
    status |= pgpERROR;
    status |= pgpMISSINGKEY;
  }
  
  if(signonly)
    output = "-----BEGIN PGP SIGNED MESSAGE-----\n\n" + input + "\n" + output;

  return status;
}

int 
KpgpBase5::decrypt(const char *passphrase)
{  
  QString in = "";
  output = "";

  status = run("pgpv -f +batchmode=1", passphrase);
  if(status == pgpRUN_ERR) return status;

  // lets parse the returned information.
  int index;

  index = info.find("Cannot decrypt message");
  if(index != -1)
  {
    //debug("message is encrypted");
    status |= pgpENCRYPTED;

    // ok. we have an encrypted message. Is the passphrase bad,
    // or do we not have the secret key?
    if(info.find("Need a pass phrase") != -1)
    {
      if(passphrase != 0) 
      {
	errMsg = i18n("Bad pass Phrase; couldn't decrypt");
	debug("KpgpBase: passphrase is bad");
	status |= pgpBADPHRASE;
	status |= pgpERROR;
      }
    }
    else
    {
      // we don't have the secret key
      status |= pgpNO_SEC_KEY;
      status |= pgpERROR;
      errMsg = i18n("Do not have the secret key for this message");
      debug("KpgpBase: no secret key for this message");
    }
    // check for persons
    index = info.find("can only be decrypted by:");
    if(index != -1) 
    {
      index = info.find("\n",index);
      int end = info.find("\n\n",index);
	 
      recipients.clear();
      int index2;
      while( (index2 = info.find("\n",index+1)) <= end )
      {
	QString item = info.mid(index+1,index2-index-1);
	item.stripWhiteSpace();
	recipients.append(item);
	index = index2;
      }
    }
  } 
  index = info.find("Good signature");
  if(index != -1)
  {
    //debug("good signature");
    status |= pgpSIGNED;
    status |= pgpGOODSIG;
 
    // get key ID of signer
    index = info.find("Key ID ");
    int index2 = info.find(",",index);
    signatureID = info.mid(index+7,index2-index-8);

    // get signer
    index = info.find("\"",index);
    index2 = info.find("\"", index+1);
    signature = info.mid(index+1, index2-index-1);
  }
  index = info.find("BAD signature");
  if(index != -1)
  {
    //debug("BAD signature");
    status |= pgpSIGNED;
    status |= pgpERROR;

    // get key ID of signer
    index = info.find("Key ID ");
    int index2 = info.find(",",index);
    signatureID = info.mid(index+7,index2-index-8);

    // get signer
    index = info.find("\"",index);
    index2 = info.find("\"", index+1);
    signature = info.mid(index+1, index2-index-1);
  }
  index = info.find("Signature by unknown key");
  if(index != -1)
  {
    index = info.find("keyid: ",index);
    int index2 = info.find("\n",index);
    signatureID = info.mid(index+7,index2-index-7);
    signature = "unknown key ID " + signatureID + " ";
    // FIXME: not a very good solution...
    status |= pgpSIGNED;
    status |= pgpGOODSIG;
  }

  //debug("status = %d",status);
  return status;
}

QStrList
KpgpBase5::pubKeys()
{
  int index,index2;
	
  status = run("pgpk -l");
  if(status == pgpRUN_ERR) return 0;

  // now we need to parse the output
  QStrList publicKeys;
  index = output.find("\n",1)+1; // skip first to "\n"
  while( (index = output.find("\n",index)) != -1)
  {
    //parse line
    QString line;
    if( (index2 = output.find("\n",index+1)) != -1)
    {
      int index3 = output.find("uid ",index);
		    
      if( (index3 <index2) && (index3 != -1) )
      {
	line = output.mid(index3+5,index2-index3-5);
	line = line.lower();
      }
      if(!line.isEmpty())
      {
	//debug("KpgpBase: found key for %s.",(const char *)line);
	publicKeys.append(line);
      }
    }
    index = index2;
  }
  //debug("finished reading keys");
  return publicKeys;
}     

QString KpgpBase5::getAsciiPublicKey(QString _person) {
  QString toexec;
  toexec.sprintf("pgpk -xa \"%s\"", _person.data());

  status = run(toexec.data());
  if(status == pgpRUN_ERR) return QStringnull();

  return output;
}

int
KpgpBase5::signKey(const char *key, const char *passphrase)
{
  QString cmd;
  
  if(passphrase == 0) return false;
  
  cmd = "pgpk -f +batchmode=1";
  cmd += key;
  cmd += addUserId();
  status = run(cmd, passphrase);

  return status;
}

int KpgpBase::runGeneric(const char *cmd)
{
  char str[1024] = "\0";
  int status;
  int pin[2], pout[2], perr[2];
  pid_t child_pid, rc;

  pipe(pin);
  pipe(pout);
  pipe(perr);

  info = "";
  output = "";

	errno = 0;
  if(!(child_pid = fork()))
  {
    /*We're the child.*/
    close(pin[1]);
    dup2(pin[0], 0);
    close(pin[0]);
  	
    close(pout[0]);
    dup2(pout[1], 1);
    close(pout[1]);

    close(perr[0]);
    dup2(perr[1], 2);
    close(perr[1]);

    execl("/bin/sh", "sh", "-c", cmd,  NULL);
    _exit(127);
  }

  /*Only get here if we're the parent.*/
  close(pin[0]);
  close(pout[1]);
  close(perr[1]);

  time_t start;
  struct timeval tmo;
  fd_set fds,wfds,efds;
	int rFlgs,wFlgs,eFlgs;
  	
	//our select timeout is 1 second
  tmo.tv_usec = 0;
  tmo.tv_sec = 1;

	//initialize all selection vectors
  FD_ZERO (&fds);
  FD_ZERO (&wfds);
  FD_ZERO (&efds);

//	debug("Start Writing to GPG %d",input.length());
	QString writebuffer;
	//we only use a copy of the buffer because
	//if we didnt have the passphrase on the first
	//try our buffer content will be gone
	if(input.isEmpty()) writebuffer = "\n";
	else writebuffer = input;
  	
//	debug("Setting non blocking IO");
	//set the non blocking flag and remember the original settings
  rFlgs = fcntl (pout[0],F_GETFL,0);
  wFlgs = fcntl (pin[1],F_GETFL,0);
  eFlgs = fcntl (perr[0],F_GETFL,0);
  fcntl (pin[1],F_SETFL,wFlgs | O_NONBLOCK);
  fcntl (pout[0],F_SETFL,rFlgs | O_NONBLOCK);
  fcntl (perr[0],F_SETFL,eFlgs | O_NONBLOCK);
  	
	bool wResult = true; //set to false by the progress bar if we press cancel
	bool rEOF,wEOF,eEOF; //indicating which streams have reached EOF
	rEOF = false;
	wEOF = false;
	eEOF = false;
  	
	//get the max. value of the file descriptors
	int max = 0;
	max = perr[0];
	if(pout[0] > max) max = pout[0];
	if(pin[1] > max) max = pin[1];
  	
	emit signalProgressNotify(mpSTART,0,"Calling GPG...",wResult);
	start = time(0);
	//while still a stream alive and cancel not pressed
  while(wResult && (!rEOF || !wEOF || !eEOF))
  {
		int i; // our universal return value
  	
		//set the read file descriptors
	  FD_ZERO (&fds);
	  FD_SET (pout[0],&fds);
	  FD_SET (perr[0],&fds);
  	
		//set the write file descriptors
	  FD_ZERO (&wfds);
		if(!wEOF)
		{
		  FD_SET (pin[1],&wfds);//only set if not already closed
	  	FD_SET (pin[1],&efds);
		}

		//we want to get error reports from those		
	  FD_ZERO (&efds);
  	FD_SET (pout[0],&efds);
  	FD_SET (perr[0],&efds);
  		
		//reset our timeout value
		tmo.tv_sec = 1;
		tmo.tv_usec = 0;
  		
		//do select
		i = select (max+1,&fds,&wfds,&efds,&tmo);
   		
 		//prints the results of select for debugging
 		debug("select returned: %d rfds: %d wfds: %d efds: %d errno: %d %s",i,
 			FD_ISSET(pout[0],&fds) + FD_ISSET(perr[0],&fds),
 			FD_ISSET(pin[1],&wfds),
 			FD_ISSET(pin[1],&efds) + FD_ISSET(pout[0],&efds) + FD_ISSET(perr[0],&efds),
 			errno,strerror(errno));
		
		//reset our global flag
		brokenPipe = false;
		  		
		//if writeable
		if(FD_ISSET(pin[1],&wfds))
		{
//			debug("Attempting write with %d",writebuffer.length());
      i = write(pin[1], writebuffer.data(), writebuffer.length());

      if(i > 0) // if succesfull write
      {
        writebuffer = writebuffer.right(writebuffer.length() - i);
        debug("Wrote %d remaining %d",i,writebuffer.length());
        if(writebuffer.isEmpty()) //if we said all we had to say
        {
  			  close(pin[1]);
  			  debug("Closing Pipe to GPG");
  			  wEOF = true;
  			}
  		} else {
  			debug("------------------------------------------------------------------Write failed %d",i);
  		}
  		if( brokenPipe )
  		{
  			debug("Pipe signal recognized");
			  close(pin[1]);
			  debug("Closing Pipe to GPG");
  			wEOF = true;
  			brokenPipe = false;
  		}

    } else {
      debug("Remaining %d",writebuffer.length());
    }
  		
		//if data waiting to read from stdout
		if(FD_ISSET(pout[0],&fds)) {
//			debug("Attempting read with 1023");
      i=read(pout[0],str,1023);
			if(i > 0) //if read succesfull
			{
        str[i] ='\0';
        output += str;

//        debug("Read %d[%d]",output.length(),i);
  		} else {
//  			debug("Read failed %d",i);
  			if(i == 0) // if EOF close stream and remember EOF
  			{
  				rEOF = true;
  				close(pout[0]);
  			}
      }
		} else {
//      debug("Read %d",output.length());
		}
  		
		//if data waiting to read from stderr
		if(FD_ISSET(perr[0],&fds)) {
//			debug("Attempting read with 1023");
      i=read(perr[0],str,1023);

			if(i > 0) //if read succesfull
			{
        str[i] ='\0';
        info += str;

//        debug("Error %d[%d]",info.length(),i);
//        debug("-> '%s'",str);
  		} else {
//  			debug("Error failed %d",i);
  			if(i == 0) // if EOF close stream and remember EOF
  			{
  				eEOF = true;
  				close(perr[0]);
  			}
      }
		} else {
//      debug("Error %d",info.length());
		}
		emit signalProgressNotify(mpTIMEOUT,time(0) - start,NULL,wResult);
  }

	//restore the blocking status ?? (arent they closed already)
  fcntl (pin[1],F_SETFL,wFlgs);
  fcntl (pout[0],F_SETFL,rFlgs);
  fcntl (perr[0],F_SETFL,eFlgs);
	debug("Stop Writing to GPG");

  // we don't want a zombie, do we?  ;-)
	if(!wResult) {
		debug("Kill Child");
    kill( child_pid, SIGHUP);
	}
	debug("Waiting for Child");
  rc = waitpid(0/*child_pid*/, &status, 0);
  	
	emit signalProgressNotify(mpSTOP,0,"Calling GPG...",wResult);
	return pgpOK;
}
